
Path ( . ) 

LayerName CWP
RGB 46 132 22
Symbolic? n
Filled 00 7e 7e 66 66 7e 7e 00
AltFilled n
MinDimensions 0
StreamData 41 0

LayerName CWN
RGB 137 124 120
Symbolic? n
Filled 00 7e 7e 66 66 7e 7e 00 outline
AltFilled n
MinDimensions 0
StreamData 42 0

LayerName CSP
RGB 255 170 174
Symbolic? n
Filled 40 20 10 08 04 02 01 80 outline
AltFilled n
MinDimensions 0
StreamData 44 0

LayerName CSN
RGB 61 211 186
Symbolic? n
Filled 40 20 10 08 04 02 01 80 outline
AltFilled n
MinDimensions 0
StreamData 45 0

LayerName CAA
RGB 254 255 1
Symbolic? n
Filled 01 82 44 28 10 28 44 82 outline
AltFilled n
MinDimensions 0
StreamData 43 0

LayerName CPG
RGB 255 85 82
Symbolic? n
Filled 18 3c 7e e7 e7 7e 3c 18 outline
AltFilled n
MinDimensions 0
StreamData 46 0

LayerName CMF
RGB 0 174 253
Symbolic? n
Filled aa 55 aa 55 aa 55 aa 55 outline
AltFilled n
MinDimensions 0
StreamData 49 0

LayerName CCP
RGB 233 151 218
Symbolic? n
Filled y
AltFilled n
MinDimensions 0
StreamData 47 0

LayerName CCA
RGB 253 253 65
Symbolic? n
Filled y
AltFilled y
MinDimensions 0
StreamData 48 0

LayerName CMS
RGB 205 0 205
Symbolic? n
Filled 2a 15 8a 45 a2 51 a8 54 outline
AltFilled n
MinDimensions 0
StreamData 51 0

LayerName CVA
RGB 240 218 181
Symbolic? n
Filled y
AltFilled n
MinDimensions 0
StreamData 50 0

LayerName COG
RGB 255 250 107
Symbolic? n
Filled n
AltFilled n
MinDimensions 0
StreamData 52 0
 

Highlighting 255 255 255
Background 0 0 0
MenuText 11
MenuSelect 9
MenuHighlighting 1
MenuPrompt 5
MoreText 255 255 255
FineGrid 100 100 100
CoarseGrid 180 180 180
InstanceBox 10
InstanceName 3
InstanceSize 4
GridSpacing 1
GridStyle 204
GridOnBottom
RoundFlashSides 40
AltDevice kicout.ps
AltResolution 150
AltFormat p
Snapping 1
FineViewportOnSide
SplitScreen
FontName 
BeepVolume 0

